﻿using PalmSens;
using PalmSens.Comm;
using PalmSens.Data;
using PalmSens.DataFiles;
using PalmSens.Devices;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace PSSDKInternalStorageExample
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
            DiscoverConnectedDevices();
        }

        /// <summary>
        /// The connected PalmSens & EmStat devices
        /// </summary>
        private Device[] _connectedDevices = new Device[0];

        /// <summary>
        /// The contents of the current directory on the internal storage
        /// </summary>
        private List<DeviceFile> _contents = new List<DeviceFile>();

        /// <summary>
        /// The measuring
        /// </summary>
        private bool _retreivingMeasurementFromStorage;

        /// <summary>
        /// Discovers the connected PalmSens & EmStat devices and adds them to the combobox control.
        /// </summary>
        private void DiscoverConnectedDevices()
        {
            cmbDevices.Items.Clear();
            _connectedDevices = psCommSimpleWinForms.ConnectedDevices; //Discover connected devices

            foreach (Device d in _connectedDevices)
                cmbDevices.Items.Add(d.ToString()); //Add connected devices to control

            int nDevices = cmbDevices.Items.Count;
            cmbDevices.SelectedIndex = nDevices > 0 ? 0 : -1;
            lbConsole.Items.Add($"Found {nDevices} device(s).");

            btnConnect.Enabled = nDevices > 0;
        }

        /// <summary>
        /// Handles the Click event of the btnConnect control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void btnConnect_Click(object sender, EventArgs e)
        {
            if (!psCommSimpleWinForms.Connected) //Determine whether a device is currently connected
            {
                if (cmbDevices.SelectedIndex == -1)
                    return;

                try
                {
                    //Connect to the device selected in the devices combobox control
                    psCommSimpleWinForms.Connect(_connectedDevices[cmbDevices.SelectedIndex]);
                    lbConsole.Items.Add($"Connected to {psCommSimpleWinForms.ConnectedDevice.ToString()}");
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            else
            {
                psCommSimpleWinForms.Disconnect(); //Disconnect from the connected device
                lbConsole.Items.Add("Disconnected");
            }

            //Update UI based on connection status
            cmbDevices.Enabled = !psCommSimpleWinForms.Connected;
            btnRefresh.Enabled = !psCommSimpleWinForms.Connected;
            btnConnect.Text = psCommSimpleWinForms.Connected ? "Disconnect" : "Connect";
            btnListFiles.Enabled = psCommSimpleWinForms.Connected && psCommSimpleWinForms.Capabilities.SupportsStorage;

            if (psCommSimpleWinForms.Connected && psCommSimpleWinForms.Capabilities.SupportsStorage)
                GetSpaceUsed();
            else if (psCommSimpleWinForms.Connected)
                lbConsole.Items.Add("The connected device does not support internal storage features");
        }

        /// <summary>
        /// Gets the % of space used on the device.
        /// </summary>
        /// <exception cref="System.NotImplementedException"></exception>
        private void GetSpaceUsed()
        {
            int total = (int)psCommSimpleWinForms.Comm.ClientConnection.GetDeviceSize();
            int free = (int)psCommSimpleWinForms.Comm.ClientConnection.GetDeviceFree();
            double perc = Math.Min((total - free) / (double)total, 1.0);
            lbConsole.Items.Add($"{perc.ToString("F2")}% of storage space used.");
        }

        /// <summary>
        /// Handles the Click event of the btnRefresh control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void btnRefresh_Click(object sender, EventArgs e)
        {
            DiscoverConnectedDevices(); //Add connected devices to the devices combobox control
        }

        /// <summary>
        /// Handles the Click event of the btnListFiles control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void btnListFiles_Click(object sender, EventArgs e)
        {
            GetFiles(""); //Get files in root directory
        }

        /// <summary>
        /// Gets the files on the internal storage.
        /// </summary>
        public void GetFiles(string path)
        {
            CommManager comm = psCommSimpleWinForms.Comm;

            string pathName = string.IsNullOrEmpty(path) ? "Root" : path;
            lbConsole.Items.Add($"Reading contents from: {pathName}");

            _contents = comm.ClientConnection.GetDeviceFiles(path); //Get the contents from the specified path

            numUDTargetItem.Enabled = _contents.Count > 0;
            btnOpen.Enabled = _contents.Count > 0;
            numUDTargetItem.Value = 1;

            if (_contents.Count > 0)
            {
                int index = 1;
                foreach (DeviceFile item in _contents)
                {
                    string type = item.Type == DeviceFileType.Folder ? "<Dir>" : "<File>";
                    lbConsole.Items.Add($"{index}. {type} {item.Name}");
                    index++;
                }
                numUDTargetItem.Maximum = _contents.Count;
            }
        }

        /// <summary>
        /// Handles the Click event of the btnOpen control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void btnOpen_Click(object sender, EventArgs e)
        {
            if (numUDTargetItem.Value < 1 || numUDTargetItem.Value > _contents.Count)
            {
                lbConsole.Items.Add("The item at the specified position does not exist");
                return;
            }

            DeviceFile item = _contents[(int)numUDTargetItem.Value - 1];

            if (item.Type == DeviceFileType.Folder)
            {
                GetFiles(item.Dir + "\\" + item.Name);
            }
            else
            {
                Measurement m = GetMeasurementFromInternalStorage(item, psCommSimpleWinForms.Comm);
            }
        }

        private Measurement GetMeasurementFromInternalStorage(DeviceFile file, CommManager comm)
        {
            //Load the actual raw data of the measurement into a DeviceFile
            DeviceFile rawData = comm.ClientConnection.GetDeviceFile(file.Dir + "/" + file.Name); 

            //Split the raw data for the method en the measurement
            int iEndChar = rawData.Content.IndexOf('*');
            if (iEndChar < 0)
                return null;

            string methodStr = rawData.Content.Substring(0, iEndChar);
            string dataStr = rawData.Content.Substring(iEndChar + 1);

            if (dataStr.Length == 0)
                return null;

            //Get the Method from the raw data
            Method m;
            using (MemoryStream memStream = new MemoryStream(Encoding.UTF8.GetBytes(methodStr)))
            using (StreamReader fileStream = new StreamReader(memStream))
                m = MethodFile2.FromStream(fileStream);

            if (!m.SupportsDeviceStorage)
                return null;

            //Get the Measurement from the raw data
            PalmSens.Comm.Buffer buff = new OfflineBuffer(dataStr);

            //Check the version and convert data in buffer is nescessary
            CheckVersion(ref buff, comm);

            ActiveMeasurement meas = m.ReceiveMeasurement(comm.ClientConnection, -1, MeasType.Overlay, buff);
            meas.EndMeasurement += Meas_EndMeasurement;

            _retreivingMeasurementFromStorage = true;
            int lastCharCnt;
            int charCntSameCnt = 0;

            //Parse raw measurement data
            while (_retreivingMeasurementFromStorage)
            {
                lastCharCnt = buff.CharsInBuffer;
                meas.ProcessData();

                if (lastCharCnt == buff.CharsInBuffer)
                {
                    if (charCntSameCnt++ > 5)
                        break; //We did not process any data last iterations, so this file is probably invalid, but still output what we have so far.
                }
                else
                    charCntSameCnt = 0;
            }

            meas.EndMeasurement -= Meas_EndMeasurement;

            return meas;
        }

        /// <summary>
        /// Checks the version of the measurement in the Buffer and converts it if necessary.
        /// </summary>
        /// <param name="buff">The buff.</param>
        /// <param name="comm">The comm.</param>
        protected void CheckVersion(ref PalmSens.Comm.Buffer buff, CommManager comm)
        {
            double fwVersionMeasurement;
            if ((char)buff.Peek() != 't')
            {
                fwVersionMeasurement = 1.5; //1.5 is the last fw version that did not have (or need) the fw version stored
            }
            else
            {
                string versionString = buff.Pop(buff.IndexOf('*') + 1).Substring(1);
                fwVersionMeasurement = ClientConnection.ParseVersion(versionString);
            }

            if (comm.ClientConnection.Capabilities.MinFirmwareVersionRequired - fwVersionMeasurement > 0.001)
            {
                buff = comm.ClientConnection.ConvertMeasurementVersion(fwVersionMeasurement, buff);
            }
        }

        private void Meas_EndMeasurement(object sender, EventArgs e)
        {
            _retreivingMeasurementFromStorage = true;
        }
    }
}
